package polyglot.ext.atomjava;

import polyglot.lex.Lexer;
import polyglot.ext.atomjava.parse.Lexer_c;
import polyglot.ext.atomjava.parse.Grm;
import polyglot.ext.atomjava.ast.*;
import polyglot.ext.atomjava.types.*;
import polyglot.ext.atomjava.visit.*;
import polyglot.ext.atomjava.main.*;
import polyglot.ext.atomjava.frontend.*;


import polyglot.ast.*;
import polyglot.types.*;
import polyglot.util.*;
import polyglot.visit.*;
import polyglot.frontend.*;
import polyglot.main.*;

import polyglot.frontend.Compiler;

import java.util.*;
import java.io.*;

/**
 * Extension information for atomjava extension.
 */
public class ExtensionInfo extends polyglot.ext.jl.ExtensionInfo {
  static {
    // force Topics to load
    Topics t = new Topics();
  }

  public String defaultFileExtension() {
    return "ajava";
  }

  public String compilerName() {
    return "atomjavac";
  }

  public Parser parser(Reader reader, FileSource source, ErrorQueue eq) {
    Lexer lexer = new Lexer_c(reader, source.name(), eq);
    Grm   grm   = new Grm(lexer, ts, nf, eq);
    return new CupParser(grm, source, eq);
  }

  protected NodeFactory createNodeFactory() {
    return new AtomJavaNodeFactory_c();
  }

  protected TypeSystem createTypeSystem() {
    return new AtomJavaTypeSystem_c(this);
  }

  protected Options createOptions() {
    return new AJOptions(this);
  }

  protected void initTypeSystem() {
    try {
      LoadedClassResolver lr = 
        new AJSourceClassResolver(compiler, this,
                                  getOptions().constructFullClasspath(),
                                  compiler.loader(), true);
      ts.initialize(lr, this);
    }
    catch (SemanticException e) {
      throw new InternalCompilerError
        ("Unable to initialize type system: " + e.getMessage());
    }
  }

  public void initCompiler(Compiler compiler) {
    super.initCompiler(compiler);
    if(((AJOptions)getOptions()).optimize) {
      optimizeJob = new OptimizeSourceJob(this, jobExt());
      optimizer = new NotInAtomicOptimizer(AJ_OPTIMIZE, optimizeJob);
    }
  }

  OptimizeSourceJob optimizeJob;
  NotInAtomicOptimizer optimizer;

  public boolean runToCompletion() {
    if(((AJOptions)getOptions()).optimize) {
      jobs.put(optimizeJob.source(), optimizeJob);
      worklist.add(optimizeJob);
    }

    return super.runToCompletion();
  }

  public SourceJob addJob(Source source, Node ast) {
    SourceJob job = super.addJob(source, ast);
    if(((AJOptions)getOptions()).optimize)
      optimizer.addSourceJob(job);
    return job;
  }

  public static final Pass.ID AJ_LOOP_LABELER
    = new Pass.ID("AJ-loop-labeler");  
  public static final Pass.ID AJ_EXTENDS_REWRITE 
    = new Pass.ID("AJ-extends-rewrite");
  public static final Pass.ID AJ_RELEASE_CHECK 
    = new Pass.ID("AJ-release-check");
  public static final Pass.ID AJ_INIT_CHECK
    = new Pass.ID("AJ-InitChecker");
  public static final Pass.ID AJ_OUTPUT 
    = new Pass.ID("AJ-output");
  public static final Pass.ID AJ_ATOMIC_BUILDER
    = new Pass.ID("AJ-atomic-builder");
  public static final Pass.ID AJ_CLASS_INIT_REWRITER
    = new Pass.ID("AJ-class-init-rewriter");
  public static final Pass.ID AJ_CALL_MANGLER
    = new Pass.ID("AJ-call-mangler");
  public static final Pass.ID AJ_ATOMIC_BUILDER_ALL
    = new Pass.ID("AJ-atomic-builder-barrier");
  public static final Pass.ID AJ_TYPED_UPDATE_ALL
    = new Pass.ID("AJ-Typed-update-all");
  public static final Pass.ID AJ_SIGNATURE_BUILDER
    = new Pass.ID("AJ-signature-builder");
  public static final Pass.ID AJ_TYPED_UPDATE_0
    = new Pass.ID("AJ-Typed-update-0");
  public static final Pass.ID AJ_TYPED_UPDATE_1
    = new Pass.ID("AJ-Typed-update-1");
  public static final Pass.ID AJ_TYPED_UPDATE_2
    = new Pass.ID("AJ-Typed-update-2"); 
  public static final Pass.ID AJ_PRE_TYPED_UPDATE_ALL_0
    = new Pass.ID("AJ-pre-Typed-update-all-0");
  public static final Pass.ID AJ_PRE_TYPED_UPDATE_ALL_1
    = new Pass.ID("AJ-pre-Typed-update-all-1");
  public static final Pass.ID AJ_CURRENT_THREAD_SHARER
    = new Pass.ID("AJ-current-thread-sharer");
  public static final Pass.ID AJ_OPTIMIZE
    = new Pass.ID("AJ-optimize");
  public static final Pass.ID AJ_BEFORE_OPTIMIZE_ALL
    = new Pass.ID("AJ-before-optimize-all");
  public static final Pass.ID AJ_AFTER_OPTIMIZE_ALL
    = new Pass.ID("AJ-after-optimize-all");

  public List optimizePasses(Job job) {
    List passes = new LinkedList();

    // jl global barriers
    passes.add(new GlobalBarrierPass(Pass.BUILD_TYPES_ALL, job));
    passes.add(new GlobalBarrierPass(Pass.ADD_MEMBERS_ALL, job));

    // atomjava global barriers
    passes.add(new GlobalBarrierPass(AJ_PRE_TYPED_UPDATE_ALL_0, job));

    // optimize global barriers
    passes.add(new GlobalBarrierPass(AJ_BEFORE_OPTIMIZE_ALL, job));
    passes.add(optimizer);
    passes.add(new GlobalBarrierPass(AJ_AFTER_OPTIMIZE_ALL, job));

    // atomjava global barriers
    passes.add(new GlobalBarrierPass(AJ_PRE_TYPED_UPDATE_ALL_1, job));
    passes.add(new GlobalBarrierPass(AJ_ATOMIC_BUILDER_ALL, job));

    return passes;
  }

  public List passes(Job job) {
    List ajpasses = new LinkedList();

    ajpasses.add(new VisitorPass(AJ_LOOP_LABELER, job,
				new BranchLabeler(job, ts, nf)));
    ajpasses.add(new VisitorPass(AJ_EXTENDS_REWRITE, job, 
                                 new ExtendsRewriter(job, ts, nf)));
    ajpasses.add(new GlobalBarrierPass(AJ_PRE_TYPED_UPDATE_ALL_0, job));
    ajpasses.add(new VisitorPass(AJ_TYPED_UPDATE_0, job, 
                                 new TypedUpdate(job, ts, nf)));
    
    if(((AJOptions)getOptions()).optimize) {
      ajpasses.add(new GlobalBarrierPass(AJ_BEFORE_OPTIMIZE_ALL, job));
      ajpasses.add(new GlobalBarrierPass(AJ_AFTER_OPTIMIZE_ALL, job));
    }

    ajpasses.add(new VisitorPass(AJ_CLASS_INIT_REWRITER, job, 
                                 new ClassInitRewriter(job, ts, nf)));
    ajpasses.add(new VisitorPass(AJ_CALL_MANGLER, job, 
                                 new CallMangler(job, ts, nf)));
    ajpasses.add(new VisitorPass(AJ_RELEASE_CHECK, job, 
                                 new ReleaseCheckInserter(nf)));
    ajpasses.add(new VisitorPass(AJ_SIGNATURE_BUILDER, job,
                                 new AJSignatureBuilder(job, ts, nf)));
    ajpasses.add(new GlobalBarrierPass(AJ_PRE_TYPED_UPDATE_ALL_1, job));
//     ajpasses.add(new VisitorPass(AJ_TYPED_UPDATE_1, job, 
//                                  new TypedUpdate(job, ts, nf)));
    ajpasses.add(new VisitorPass(AJ_ATOMIC_BUILDER, job, 
                                 new AtomJavaBuilder(job, ts, nf)));
    ajpasses.add(new GlobalBarrierPass(AJ_ATOMIC_BUILDER_ALL, job));
    ajpasses.add(new VisitorPass(AJ_CURRENT_THREAD_SHARER, job,
                                 new CurrentThreadSharer(job,ts,nf)));
//     ajpasses.add(new VisitorPass(AJ_TYPED_UPDATE_2, job,
//                                  new TypedUpdate(job, ts, nf)));

    List passes = super.passes(job);
    beforePass(passes, Pass.PRE_OUTPUT_ALL, ajpasses);
            
//     replacePass(passes, Pass.INIT_CHECK, 
//                 new VisitorPass(AJ_INIT_CHECK, job, new AJInitChecker(job, ts, nf)));
    
//      for(Iterator i = passes.iterator(); i.hasNext();)
//        System.out.println(i.next());

    return passes;
  }

  protected Job createJob(Node ast, Context context, Job outer, 
                          NodeVisitor v, Pass.ID id) {
    return new SingleVisitorJob(this, jobExt(), ast, context, outer, v, id);
  }

  public Job spawnJob(Context c, Node ast, Job outerJob, 
                      NodeVisitor v, Pass.ID id) {
    Job j = createJob(ast, c, outerJob, v, id);

    if (Report.should_report(Report.frontend, 1))
      Report.report(1, this +" spawning " + j);

    // Run all the passes
    runAllPasses(j);

    // Return the job.  The caller can check the result with j.status().
    return j;
  }

  /* Pass Dependencies:
        the dependency relation is a happens-before dependency.  
        <a,b> s.t. a depends on b 
  */
  private HashMap<Pass.ID,Pass.ID> pass_dependencies = 
    new HashMap<Pass.ID,Pass.ID>();

  private void addDependencies(List newPasses, Pass.ID dependency) {
    for (Iterator j = newPasses.iterator(); j.hasNext(); ) {
      Pass p = (Pass) j.next();
      pass_dependencies.put(p.id(), dependency);
    }
  }

  public void beforePassDependent(List passes, Pass.ID id, 
                                  List newPasses, Pass.ID dependency) {
    boolean earlier = false;

    for (ListIterator i = passes.listIterator(); i.hasNext();) {
      Pass p = (Pass) i.next();

      if(p.id() == dependency && p.id() != id)  
        earlier = true;
      
      if(p.id() == id && earlier) {
        super.beforePass(passes, id, newPasses);
        return; // so we don't get a comodification error
      }
      
      if(p.id() == id && !earlier)
        throw new InternalCompilerError
          ("Inserted passes at " + id + " can not depend upon later Pass " + dependency);
    }
    /* add the dependencies if necessary */
    addDependencies(newPasses, dependency);
  }

  public void afterPassDependent(List passes, Pass.ID id,
                                 List newPasses, Pass.ID dependency) {
    boolean earlier = false;

    for (ListIterator i = passes.listIterator(); i.hasNext();) {
      Pass p = (Pass) i.next();

      if(p.id() == dependency)
        earlier = true;
      
      if(p.id() == id && earlier) {
        super.afterPass(passes, id, newPasses);
        return; // so we don't get a comodification error
      }
      
      if(p.id() == id && !earlier)
        throw new InternalCompilerError
          ("Inserted passes at " + id + " can not depend upon later Pass " + dependency);
    }
    /* add dependencies if necssary */
    addDependencies(newPasses, dependency);
  }

  public void replacePass(List passes, Pass.ID id, List newPasses) {
    if(pass_dependencies.containsValue(id))
      throw new InternalCompilerError("Pass " + id + " is a dependency.");
    super.replacePass(passes, id, newPasses);
  }

  public void removePass(List passes, Pass.ID id) {
    if(pass_dependencies.containsValue(id))
      throw new InternalCompilerError("Pass " + id + " is a dependency.");
    super.removePass(passes, id);
  }

  public void beforePassDependent(List passes, Pass.ID id, 
                                  Pass pass, Pass.ID dependency) {
    beforePassDependent(passes, id, Collections.singletonList(pass),
                        dependency);
  }
  
  public void afterPassDependent(List passes, Pass.ID id, 
                                 Pass pass, Pass.ID dependency) {
    afterPassDependent(passes, id, Collections.singletonList(pass),
              dependency);
  }  

}
