package polyglot.ext.atomjava.ast;

import polyglot.ext.atomjava.types.AtomJavaContext;
import polyglot.ext.atomjava.types.AtomJavaTypeSystem;
import polyglot.ext.atomjava.visit.*;
import polyglot.ext.atomjava.main.*;
import polyglot.ext.atomjava.util.AJS;
import polyglot.ext.jl.ast.MethodDecl_c;
import polyglot.ast.*;
import polyglot.util.*;
import polyglot.types.*;
import polyglot.visit.*;

import java.util.*;

public abstract class AJAbstractMethodDecl
  extends MethodDecl_c 
  implements AtomJavaBuild {

  public AJAbstractMethodDecl(Position pos, Flags flags, 
                              TypeNode returnType, String name, 
                              List formals, List throwTypes, Block body) {
    super(pos, flags, returnType, name, formals, throwTypes, body);
  }
  
  private boolean needsReleaseCheck() {
    // conservative -- just need to break loops in CFG
    // easy case is returning false for leaf procedures
    // note: assuming Java forbids cycles in constructor calls
    return true;
  }

  public AJAbstractMethodDecl convertSynchronized(AtomJavaContext c, 
                                                  AtomJavaNodeFactory nf, 
                                                  AtomJavaTypeSystem ts) {
    if(!flags().isSynchronized())
      return this;

    // Dan: Must build AtomJava node while synchronized still uses translate!!
    Position p = Position.COMPILER_GENERATED;

    Expr e = flags().isStatic() 
      ? nf.ClassLit(p, nf.CanonicalTypeNode(p,c.currentClass())).type(ts.Class())
      : nf.Special(p, Special.THIS, null).type(c.currentClass());

    Synchronized s = nf.Synchronized(p, e, nf.Block(c, body().statements()));

    MethodDecl md = flags(flags().clearSynchronized());
    md = (MethodDecl) md.body(nf.Block(c, AJS.List(s)));
    MethodInstance mi = md.methodInstance();
    mi = mi.flags(md.flags());
    md = md.methodInstance(mi);
    return (AJAbstractMethodDecl) md;
  }

  protected boolean isMain(AtomJavaBuilder ab) {
    MethodInstance mi = methodInstance();
    
    if(!(mi.name().equals("main") && 
         mi.returnType().isVoid() &&
         (Flags.PUBLIC.set(Flags.STATIC).equals(flags)) &&
         formals().size() == 1))
      return false;
    
    Formal formal = (Formal)formals().get(0);
    
    if(!formal.declType().isArray())
      return false;
    
    Type t = formal.declType().toArray().base();
    
    if(!t.isClass() || !t.toClass().fullName().equals("java.lang.String"))
      return false;
    
    return true;
  }

  protected Node build_amain(AtomJavaBuilder ab) {
    
    AtomJavaContext c = ab.context();
    AtomJavaNodeFactory nf = ab.nodeFactory();
    AtomJavaTypeSystem ts = ab.typeSystem();

    Block b = nf.Block(c, Collections.EMPTY_LIST);    
    LocalDecl fargs = 
      nf.LocalDecl(c, Flags.FINAL, 
                   nf.CanonicalTypeNode(c, ts.arrayOf(ts.String())),
                   "fargs", nf.Local(c, ((Formal)formals.get(0))));
    b = b.append(fargs);
    CanonicalTypeNode amain = nf.CanonicalTypeNode(c, ts.AMain());
    Call created = nf.Call(c, amain, "createdMainThread", 
                           Collections.EMPTY_LIST);
    Call main = nf.Call(c, nf.CanonicalTypeNode(c, c.currentClass()),
                        "main", AJS.List(nf.Local(c, fargs)));
    MethodDecl run = nf.MethodDecl(c, Flags.PUBLIC, ts.Void(),
                                   "run", Collections.EMPTY_LIST,
                                   AJS.List(nf.Eval(c, main)));    
    ClassBody cb = nf.ClassBody(c, AJS.List(run));
    New n = nf.New(c, null, nf.CanonicalTypeNode(c, ts.AThread()),
                   Collections.EMPTY_LIST, cb);
    Call create = nf.Call(c, amain, "createMainThread",
                          AJS.List(n));
    If i = nf.If(c, nf.Unary(c, Unary.NOT, created), 
                 nf.Eval(c, create), body);
    b = b.append(i);
    return nf.MethodDecl(c, flags, returnType, name,
                         formals, throwTypes, b);
  }

  public NodeVisitor buildEnter(AtomJavaBuilder ab) {
    return ab;
  }

  public abstract Node build(Node parent, AtomJavaBuilder ab);
  
}

