package polyglot.ext.atomjava.ast;

import polyglot.ext.atomjava.visit.AtomJavaBuilder;
import polyglot.ext.jl.ast.FieldAssign_c;
import polyglot.util.CodeWriter;
import polyglot.visit.Translator;
import polyglot.ast.*;
import polyglot.util.Position;
import polyglot.ast.Receiver;
import polyglot.ast.Binary;
import polyglot.ast.Variable;
import polyglot.types.*;
import polyglot.visit.*;
import polyglot.ext.atomjava.frontend.*;
import polyglot.ext.atomjava.main.AJOptions;
import polyglot.ext.atomjava.types.AtomJavaContext;
import polyglot.ext.atomjava.types.AtomJavaTypeSystem;
import polyglot.ext.atomjava.util.AJS;

import java.util.*;

public class AtomJavaFieldAssign_c 
  extends FieldAssign_c 
  implements AtomJavaBuild {

  public AtomJavaFieldAssign_c(Position pos, Field left, 
                               Assign.Operator op, Expr right) {
    super(pos, left, op, right);
  }

  public NodeVisitor buildEnter(AtomJavaBuilder ab) {
    return ab;
  }

  public Node build(Node parent, AtomJavaBuilder ab) {
    // TODO: for now getters and setters don't exist for final fields, 
    // eventually we will need to work around setting final fields in
    // either static initialization or constructors
    if(((Field)left).flags().isFinal())
      return buildWeak(parent, ab);

    AJOptions options = (AJOptions) ab.job().extensionInfo().getOptions();

    if(options.optimize && 
       !NotInAtomicOptimizer.fieldReadInAtomic(((Field)left).fieldInstance()) &&
       !NotInAtomicOptimizer.fieldWrittenInAtomic(((Field)left).fieldInstance())) {
      System.out.println("OPTIMIZING WRITE << !read_inatomic && !written_inatomic >>: " + ((Field)left).fieldInstance());
      return buildWeak(parent, ab);
    }

    if(options.weak && !ab.context().inAtomicContext())
      return buildWeak(parent, ab);

    return buildStrong(parent, ab);
  }

  public Node buildWeak(Node parent, AtomJavaBuilder ab) {
    return this;
  }

  public Node buildStrong(Node parent, AtomJavaBuilder ab) {
    AtomJavaContext     ajc = ab.context();
    AtomJavaNodeFactory nf  = ab.nodeFactory();
    AtomJavaTypeSystem  ts  = ab.typeSystem();

    Field field = (Field) left;
    Receiver e1 = field.target();
    Expr target = (e1 instanceof Expr) ? (Expr)e1 : nf.NullLit(ajc);

    List call_args;

    if((left.type().isChar() || 
       left.type().isByte() || 
       left.type().isShort()) &&
       !ts.equals(left.type(), right.type()) &&
       !ts.isImplicitCastValid(right.type(), left.type())) {
      call_args = 
        AJS.List(target, 
                 nf.Cast(ajc, nf.CanonicalTypeNode(ajc, left.type()), right));
    } else call_args = AJS.List(target,right);

    // if the type of the class containing the field is an innerClass
    // then look for the first outer class that can have static members
    // and make a call to that class
    ClassType ct = (ClassType) e1.type();
    String mangled_name = field.name();
    for(; ct.isInnerClass(); ct = ct.outer())
	mangled_name = ct.name() + "_" + mangled_name;


    if(op == ASSIGN)
      return nf.Call(ajc, nf.CanonicalTypeNode(ajc, ct),
                     ajc.fieldSetterName(mangled_name), call_args);

    return nf.Call(ajc, null, call_args,
                   ajc.makeFieldOperatorMethod(field, op, ab));
  }
}
