package polyglot.ext.atomjava.ast;

import polyglot.ext.jl.ast.Field_c;
import polyglot.util.CodeWriter;
import polyglot.visit.Translator;
import polyglot.types.*;
import polyglot.ast.*;
import polyglot.ext.atomjava.types.AtomJavaContext;
import polyglot.ext.atomjava.types.AtomJavaTypeSystem;
import polyglot.util.*;
import polyglot.ext.atomjava.util.*;
import polyglot.ext.atomjava.frontend.*;
import polyglot.ext.atomjava.main.AJOptions;
import polyglot.ext.atomjava.visit.AtomJavaBuilder;
import polyglot.visit.*;
import java.util.List;

public class AtomJavaField_c 
  extends Field_c 
  implements AtomJavaBuild {

  public AtomJavaField_c(Position pos, Receiver target, String name) {
    super(pos, target, name);
  }

  public NodeVisitor buildEnter(AtomJavaBuilder ab) {
    return ab;
  }

  public Node build(Node parent, AtomJavaBuilder ab) {
    AJOptions options = (AJOptions) ab.job().extensionInfo().getOptions();

    if(options.optimize && 
       !NotInAtomicOptimizer.fieldReadInAtomic(fieldInstance()) &&
       !NotInAtomicOptimizer.fieldWrittenInAtomic(fieldInstance())) {
      System.out.println("OPTIMIZING READ << !read_inatomic && !written_inatomic >>: " + fieldInstance());
      return buildWeak(parent, ab);
    }

    if(options.optimize && 
       NotInAtomicOptimizer.fieldReadInAtomic(fieldInstance()) &&
       !NotInAtomicOptimizer.fieldWrittenInAtomic(fieldInstance()) &&
       !ab.context().inAtomicContext()) {
      System.out.println("OPTIMIZING READ << read_inatomic && !written_inatomic && !inatomic >>: " + fieldInstance());
      return buildWeak(parent, ab);
    }

    if(options.weak && !ab.context().inAtomicContext())
      return buildWeak(parent, ab);

    return buildStrong(parent, ab);
  }

  public Node buildWeak(Node parent, AtomJavaBuilder ab) {
    return this;
  }

  public Node buildStrong(Node parent, AtomJavaBuilder ab) {
    AtomJavaContext ajc = (AtomJavaContext) ab.context();
    AtomJavaNodeFactory nf = (AtomJavaNodeFactory) ab.nodeFactory();
    AtomJavaTypeSystem ts = (AtomJavaTypeSystem) ab.typeSystem();

    /* TO DO: verify that this isn't in java.lang */

    /* ignore array's length field && ignore final fields */
    if(target.type().isArray() || fi.flags().isFinal())
      return this;

    // if the type of the class containing the field is an innerClass
    // then look for the first outer class that can have static members
    // and make a call to that class
    ClassType ct = (ClassType) target.type();
    String mangled_name = name;
    for(; ct.isInnerClass(); ct = ct.outer())
	mangled_name = ct.name() + "_" + mangled_name;

    Expr e = (target instanceof Expr) ? (Expr) target : nf.NullLit(ajc);
    
    return nf.Call(ajc, nf.CanonicalTypeNode(position, ct), 
                   ajc.fieldGetterName(mangled_name), AJS.List(e));
  }
}
