package polyglot.ext.atomjava.ast;

import polyglot.ext.atomjava.types.*;
import polyglot.ext.atomjava.visit.*;
import polyglot.ext.atomjava.util.*;
import polyglot.ext.jl.ast.New_c;
import polyglot.util.*;
import polyglot.visit.*;
import polyglot.types.*;
import polyglot.ast.*;
import java.util.*;

public class AtomJavaNew_c extends New_c implements AtomJavaBuild {

  public AtomJavaNew_c(Position pos, Expr outer, TypeNode objectType, 
                       List args, ClassBody body) {
    super(pos,outer,objectType,args,body);
  }

  public NodeVisitor buildEnter(AtomJavaBuilder ab) {
    return ab;
  }

  public Node build(Node parent, AtomJavaBuilder ab) {

    AtomJavaContext     c  = ab.context();
    AtomJavaNodeFactory nf = ab.nodeFactory();
    AtomJavaTypeSystem  ts = ab.typeSystem();

    /* replace new Object() -> new AObject() */
    if(ts.equals(objectType().type(), ts.Object())) {
      New n = objectType(nf.CanonicalTypeNode(c, ts.AObject()));
      n = (New) n.type(ts.AObject());
      ConstructorInstance ci = n.constructorInstance();
      ci = ci.container(ts.AObject());
      n = n.constructorInstance(ci);
      return n;
    }

    if(ab.context().inAtomicContext())
      return buildAtomic(parent, ab);
    return buildNotAtomic(parent,ab);
  }

  private Node Failure(AtomJavaBuilder ab, AtomJavaContext c, AtomJavaNodeFactory nf, Node parent) {
    ab.errorQueue().enqueue(ErrorInfo.WARNING, "No wrapper method found for constructor " + ci.signature() + ".", position());
        
    Expr e = (qualifier() != null) ? qualifier() : nf.NullLit(c);
    Call fail = nf.Failure(c, objectType().type(), toString(), 
                           AJS.cons(e, new LinkedList(arguments())));
    if(!(parent instanceof Eval))
      return nf.Cast(c, nf.CanonicalTypeNode(c, objectType().type()), fail);
    return fail;
  }

  private Node getWrapper(Node parent, AtomJavaBuilder ab, String name) {
    AtomJavaContext     c  = ab.context();
    AtomJavaNodeFactory nf = ab.nodeFactory();
    AtomJavaTypeSystem  ts = ab.typeSystem();
    ClassType t = (ClassType)constructorInstance().container();
    try {
      /* look for wrapper */
      ClassType wt = 
        (ClassType) ts.typeForName(AtomJavaTypeSystem.RUNTIME + "wrapper." +
                                   t.fullName());

      ts.findMethod(wt, name, constructorInstance().formalTypes(), c.currentClass());
      return nf.Call(c, nf.CanonicalTypeNode(c,wt), name, arguments);
    } catch (SemanticException e) {}
    /* no wrapper found, fail */
    return Failure(ab, c, nf, parent);
  }

  public Node buildAtomic(Node parent, AtomJavaBuilder ab) {
    AtomJavaContext     c  = ab.context();
    AtomJavaNodeFactory nf = ab.nodeFactory();
    AtomJavaTypeSystem  ts = ab.typeSystem();
    
    List args = new LinkedList(arguments);
    Field f = nf.Field(c, ts.AConstructorFormal(), 
                       c.atomicConstructorFieldName());
    args.add(0, f);
    /* determination of valid atomic instantiation */
    ClassType t = (ClassType)constructorInstance().container();
    if(ts.useWrappers(t))
      return getWrapper(parent,ab,"__aj_atomicnew");
    try {
      ConstructorInstance ci = 
        ts.findConstructor(constructorInstance().container().toClass(), 
                           AJS.Expr_to_Type(args), c.currentClass());
        
      New n = (New) arguments(args);
      return n.constructorInstance(ci);
    } catch (SemanticException e) {}
      
    throw new InternalCompilerError
      ("Subtype of AObject missing atomic constructor(s)", position());
  }
  
  public Node buildNotAtomic(Node parent, AtomJavaBuilder ab) {
    AtomJavaTypeSystem  ts = ab.typeSystem();
    ClassType t = (ClassType)constructorInstance().container();
    if(ts.useWrappers(t))
      return getWrapper(parent,ab,"__aj_new");
    return this;
  }
}
