package polyglot.ext.atomjava.ast;

import polyglot.ast.*;
import polyglot.ext.jl.ast.*;
import polyglot.types.*;
import polyglot.util.*;
import java.util.*;

/**
 * NodeFactory for atomjava extension.
 */
public interface AtomJavaNodeFactory extends NodeFactory {
  AtomicBlock AtomicBlock(Position pos, Block block);
  ReleaseMaybeCheck ReleaseMaybeCheck();


    //////////////////////////////////////////////////////////////////////
   ////////                     RUNTIME AST               ///////////////
  //////////////////////////////////////////////////////////////////////

  public Try Try(Context c, Block tryBlock, List catchBlocks, 
                 Block finallyBlock);

  public Do Do(Context c, Stmt body, Expr cond);

  public While While(Context c, Expr cond, Stmt body);

  public Catch Catch(Context c, Formal formal, Block body);

  public Labeled Labeled(Context c, String label, Stmt statement);

  public Branch Branch(Context c, Branch.Kind kind, String label);

  public If If(Context c, Expr cond, Stmt consequent);

  public If If(Context c, Expr cond, Stmt consequent, Stmt alternative);

  public StringLit StringLit(Context c, String value);

  public Block Block(Context c, List statements);

  public Throw Throw(Context c, Expr expr);

  public IntLit IntLit(Context c, IntLit.Kind kind, long value);

  public Return Return(Context c, Expr expr);

  public Eval Eval(Context c, Expr expr);

  public BooleanLit BooleanLit(Context c, boolean value);

  public Conditional Conditional(Context c, Expr cond, Expr consequent, 
                                 Expr alternative);

  public NullLit NullLit(Context c);
  
  public Synchronized Synchronized(Context c, Expr expr, Block body);
  
  public MethodDecl MethodDecl(Context c, Flags flags, TypeNode returnType, 
                               String name, List formals, List throwTypes, 
                               Block body);

  public MethodDecl MethodDecl(Context c, Flags flags, Type returnType, 
                               String name, List formals, List statements);

  public MethodDecl AtomicMethodDecl(Context c, Flags flags, 
                                     TypeNode returnType, 
                                     String name, List formals, 
                                     List throwTypes, Block body);

  public ConstructorDecl ConstructorDecl(Context c, Flags flags, String name, 
                                         List formals, List throwTypes, 
                                         Block body);

  public ConstructorDecl AtomicConstructorDecl(Context c, Flags flags, 
                                                     String name, List formals, 
                                                     List throwTypes, 
                                                     Block body);
  
  public FieldDecl FieldDecl(Context c, Flags flags, TypeNode type, 
                             String name, Expr init);

  public Formal Formal(Context c, Flags flags, TypeNode type, String name);
  
  public Cast Cast(Context c, TypeNode type, Expr expr);
  
  public Call Call(Context c, Receiver target, String name,
                   List args);

  public Call Failure(Context c, Type type, String s, List exprs);
  
  public Call currentThread(Context c);
  public Field threadLocalField(Context c, String fieldName);
  public Throw doRollBack(Context c);

  public Assign Assign(Context c, Expr left, Assign.Operator op, Expr right);

  public Assign Assign(Context c, Expr left, Expr right);
    
  public Binary Binary(Context c, Expr left, 
                       Binary.Operator op, 
                       Expr right, Type t);

  public Unary Unary(Context c, Unary.Operator op, Expr expr);

  public CanonicalTypeNode CanonicalTypeNode(Context c, Type t);

  public New New(Context c, Expr outer, TypeNode objectType, 
                 List args, ClassBody body);

  public LocalDecl LocalDecl(Context c, Flags flags, TypeNode type, 
                             String name, Expr init);

  public Initializer Initializer(Context c, Flags f, Block b, 
                                 ClassType container);
  
  public ClassBody ClassBody(Context c, List members);

  public LocalDecl freshLocalDecl(Context c, Flags flags, String startName, 
                                  Type type, Expr init);

  public Lit Zero(Context c, Type type);

  //// factory for nodes with relationships ////

  public Call Call(Context c, Receiver target, List args, MethodDecl md);

  public Local Local(Context c, LocalInstance li);
  
  public Local Local(Context c, LocalDecl ld);
  
  public Local Local(Context c, Formal f);
  
  public Field Field(Context c, FieldDecl fd, Receiver r);
  
  public Field Field(Context c, FieldDecl fd);
  
  public Field Field(Context c, ClassType t, String name);

  public Field Field(Context c, ClassType t, Receiver r, String name);
}
