package polyglot.ext.atomjava.ast;

import polyglot.ast.*;
import polyglot.ext.jl.ast.*;
import polyglot.ext.jl.types.*;
import polyglot.types.*;
import polyglot.util.*;
import java.util.*;
import polyglot.ext.atomjava.util.AJS;
import polyglot.ext.atomjava.types.AtomJavaTypeSystem;

/**
 * NodeFactory for atomjava extension.
 */
public class AtomJavaNodeFactory_c extends NodeFactory_c implements AtomJavaNodeFactory {
  private Position cgp = Position.COMPILER_GENERATED;
  
  public AtomJavaNodeFactory_c() {
    super(new AtomJavaExtFactory_c());
  }
  
  public AtomicBlock AtomicBlock(Position pos, Block block) { 
    AtomicBlock ab = new AtomicBlock_c(pos, block.statements());
    ab = (AtomicBlock)ab.ext(((AtomJavaExtFactory_c)extFactory()).extAtomicBlock());	
    // 	ab = (AtomicBlock)ab.del(((AtomJavaDelFactory)delFactory()).delAtomicBlock());	
    return ab;
  }
  
  public ReleaseMaybeCheck ReleaseMaybeCheck() {
    ReleaseMaybeCheck r = new ReleaseMaybeCheck_c();
    r = (ReleaseMaybeCheck)r.ext(((AtomJavaExtFactory_c)extFactory()).extReleaseMaybeCheck());
    return r;
  }

  public Call Call(Position pos, Receiver target, String name, List args) {
    Call n = new AtomJavaCall_c(pos, target, name, args);
    n = (Call)n.ext(extFactory().extCall());
    n = (Call)n.del(delFactory().delCall());
    return n;
  }

  public MethodDecl MethodDecl(Position pos, Flags flags, TypeNode returnType, String name, List formals, List throwTypes, Block body) {
    MethodDecl n = new AtomJavaMethodDecl_c(pos, flags, returnType, name, formals, throwTypes, body);
    n = (MethodDecl)n.ext(extFactory().extMethodDecl());
    n = (MethodDecl)n.del(delFactory().delMethodDecl());
    return n;
  }

  public ConstructorCall ConstructorCall(Position pos, ConstructorCall.Kind kind, Expr outer, List args) {
    ConstructorCall n = new AtomJavaConstructorCall_c(pos, kind, outer, args);
    n = (ConstructorCall)n.ext(extFactory().extConstructorCall());
    n = (ConstructorCall)n.del(delFactory().delConstructorCall());
    return n;
  }

  public New New(Position pos, Expr outer, TypeNode objectType, List args, ClassBody body) {
    New n = new AtomJavaNew_c(pos, outer, objectType, args, body);
    n = (New)n.ext(extFactory().extNew());
    n = (New)n.del(delFactory().delNew());
    return n;
  }

  public FieldAssign FieldAssign(Position pos, Field left, Assign.Operator op, Expr right) {
    FieldAssign n = new AtomJavaFieldAssign_c(pos, left, op, right);
    n = (FieldAssign)n.ext(extFactory().extFieldAssign());
    n = (FieldAssign)n.del(delFactory().delFieldAssign());
    return n;
  }

  public Unary Unary(Position pos, Unary.Operator op, Expr expr) {
    Unary n = new AtomJavaUnary_c(pos, op, expr);
    n = (Unary)n.ext(extFactory().extUnary());
    n = (Unary)n.del(delFactory().delUnary());
    return n;
  }

  public Field Field(Position pos, Receiver target, String name) {
    Field n = new AtomJavaField_c(pos, target, name);
    n = (Field)n.ext(extFactory().extField());
    n = (Field)n.del(delFactory().delField());
    return n;
  }

  public ArrayAccess ArrayAccess(Position pos, Expr base, Expr index) {
    ArrayAccess n = new AtomJavaArrayAccess_c(pos, base, index);
    n = (ArrayAccess)n.ext(extFactory().extArrayAccess());
    n = (ArrayAccess)n.del(delFactory().delArrayAccess());
    return n;
  }

  public ArrayAccessAssign ArrayAccessAssign(Position pos, ArrayAccess left, Assign.Operator op, Expr right) {
    ArrayAccessAssign n = new AtomJavaArrayAccessAssign_c(pos, left, op, right);
    n = (ArrayAccessAssign)n.ext(extFactory().extArrayAccessAssign());
    n = (ArrayAccessAssign)n.del(delFactory().delArrayAccessAssign());
    return n;
  }

  public Synchronized Synchronized(Position pos, Expr expr, Block body) {
    Synchronized n = new AtomJavaSynchronized_c(pos, expr, body);
    n = (Synchronized)n.ext(extFactory().extSynchronized());
    n = (Synchronized)n.del(delFactory().delSynchronized());
    return n;
  }

  public Try Try(Position pos, Block tryBlock, List catchBlocks, Block finallyBlock) {
    Try n = new AtomJavaTry_c(pos, tryBlock, catchBlocks, finallyBlock);
    n = (Try)n.ext(extFactory().extTry());
    n = (Try)n.del(delFactory().delTry());
    return n;
  }

    //////////////////////////////////////////////////////////////////////
   ////////                     RUNTIME AST               ///////////////
  //////////////////////////////////////////////////////////////////////


  public Try Try(Context c, Block tryBlock, List catchBlocks, 
                 Block finallyBlock) {
    return super.Try(cgp, tryBlock, catchBlocks, finallyBlock);
  }

  public While While(Context c, Expr cond, Stmt body) {
    return super.While(cgp, cond, body); 
  }

  public Do Do(Context c, Stmt body, Expr cond) {
    return super.Do(cgp, body, cond);
  }

  public Catch Catch(Context c, Formal formal, Block body) {
    return super.Catch(cgp, formal, body);
  }

  public Labeled Labeled(Context c, String label, Stmt statement) {
    return super.Labeled(cgp, label, statement);
  }

  public Branch Branch(Context c, Branch.Kind kind, String label) {
    return super.Branch(cgp, kind, label);
  }

  public If If(Context c, Expr cond, Stmt consequent, Stmt alternative) {
    return super.If(cgp, cond, consequent, alternative);
  }
  
  public If If(Context c, Expr cond, Stmt consequent) {
      return super.If(cgp, cond, consequent);
  }

  public StringLit StringLit(Context c, String value) {
    return (StringLit) super.StringLit(cgp, value)
      .type(c.typeSystem().String());
  }

  public Block Block(Context c, List statements) {
    return super.Block(cgp, statements);
  }

  public Throw Throw(Context c, Expr expr) {
    return super.Throw(cgp, expr);
  }

  public IntLit IntLit(Context c, IntLit.Kind kind, long value) {
    IntLit i = super.IntLit(cgp, kind, value);
    if(kind == IntLit.INT)
      return (IntLit) i.type(c.typeSystem().Int());
    else //assert(kind == IntLit.LONG)
      return (IntLit) i.type(c.typeSystem().Long());    
  }
 
  public Return Return(Context c, Expr expr) {
    return super.Return(cgp, expr);
  }

  public Eval Eval(Context c, Expr expr) {
    return super.Eval(cgp, expr);
  }

  public BooleanLit BooleanLit(Context c, boolean value) {
    return (BooleanLit) super.BooleanLit(cgp, value)
      .type(c.typeSystem().Boolean());
  }

  public Conditional Conditional(Context c, Expr cond, Expr consequent, 
                                 Expr alternative) {
    try {
      /* set the conditional's type correctly */
      Type type = c.typeSystem().leastCommonAncestor(consequent.type(),
                                                     alternative.type());
      Conditional con = super.Conditional(cgp, cond, consequent, alternative);
      return (Conditional) con.type(type);
    } catch (SemanticException e) { throw new InternalCompilerError(e); }
  }

  public NullLit NullLit(Context c) {
    return (NullLit) super.NullLit(cgp).type(c.typeSystem().Null());
  }

  public Synchronized Synchronized(Context c, Expr expr, Block body) {
    return super.Synchronized(cgp, expr, body);
  }

  public MethodDecl MethodDecl(Context c, Flags flags, TypeNode returnType, 
                               String name, List formals, List throwTypes, 
                               Block body) {
    MethodDecl md = 
      super.MethodDecl(cgp, flags, returnType, name, 
                       formals, throwTypes, body);
    
    return md.methodInstance
      (new MethodInstance_c(c.typeSystem(), cgp, c.currentClass(), flags, 
                            returnType.type(), name, 
                            AJS.Formal_to_Type(formals), 
                            AJS.TypeNode_to_Type(throwTypes)));
  }

  public MethodDecl MethodDecl(Context c, Flags flags, Type returnType, 
                               String name, List formals, List statements) {
    return MethodDecl(c, flags, CanonicalTypeNode(c,returnType), name, formals,
                      Collections.EMPTY_LIST, Block(cgp, statements));
  }

  public MethodDecl AtomicMethodDecl(Context c, Flags flags, 
                                     TypeNode returnType, 
                                     String name, List formals, 
                                     List throwTypes, Block body) {
    MethodDecl md = new AtomJavaAtomicMethodDecl_c(cgp, flags, returnType,
                                                   name, formals,
                                                   throwTypes, body);
    return md.methodInstance
      (new MethodInstance_c(c.typeSystem(), cgp, c.currentClass(), flags,
                            returnType.type(), name,
                            AJS.Formal_to_Type(formals),
                            AJS.TypeNode_to_Type(throwTypes)));
  }

  public ConstructorDecl ConstructorDecl(Context c, Flags flags, String name, 
                                         List formals, List throwTypes, 
                                         Block body) {
    ConstructorDecl cd = 
      super.ConstructorDecl(cgp, flags, name, formals, throwTypes, body);

    return cd.constructorInstance
      (new ConstructorInstance_c(c.typeSystem(), cgp, c.currentClass(), 
                                 flags, AJS.Formal_to_Type(formals),
                                 AJS.TypeNode_to_Type(throwTypes)));
  }

  public ConstructorDecl AtomicConstructorDecl(Context c, Flags flags, 
                                               String name, List formals, 
                                               List throwTypes, 
                                               Block body) {
    ConstructorDecl cd = 
      new AtomJavaAtomicConstructorDecl_c(cgp, flags, name, formals, 
                                          throwTypes, body);

    return cd.constructorInstance
      (new ConstructorInstance_c(c.typeSystem(), cgp, c.currentClass(),
                                 flags, AJS.Formal_to_Type(formals),
                                 AJS.TypeNode_to_Type(throwTypes)));
  }  

  public FieldDecl FieldDecl(Context c, Flags flags, TypeNode type, String name, Expr init) {
    FieldDecl fd = super.FieldDecl(cgp, flags, type, name, init);
    return fd.type(type).fieldInstance
      (new FieldInstance_c(c.typeSystem(), fd.position(), c.currentClass(), 
                           fd.flags(), fd.declType(), fd.name()));        
  }
  
  public Formal Formal(Context c, Flags flags, TypeNode type, String name) {
    Formal f = super.Formal(cgp, flags, type, name);
    return f.type(type).localInstance
      (new LocalInstance_c(c.typeSystem(), f.position(), 
                           f.flags(), f.declType(), f.name()));
  }
    
  public Cast Cast(Context c, TypeNode type, Expr expr) {
    Cast n = super.Cast(cgp, type, expr);
    return (Cast)n.type(type.type());
  }

  public Call Call(Context c, Receiver target, String name, 
                   List args) {
    Call call = super.Call(cgp, target, name, args);

    AtomJavaTypeSystem ts = (AtomJavaTypeSystem)c.typeSystem();
    try {

      ReferenceType ref = call.isTargetImplicit()
        ? c.currentClass() : (ReferenceType) call.target().type();

      MethodInstance mi = ts.findMethod(ref, call.name(), 
                                        AJS.Expr_to_Type(call.arguments()), 
                                        //formalTypes,
                                        c.currentClass());

      call = (Call) call.type(mi.returnType());
      return call.methodInstance(mi);
    } catch(SemanticException e) { 
      throw new InternalCompilerError
        ("MethodInstance not found!: " + call, e);
    }
  }

  public Call Failure(Context c, Type type, String s, List exprs) {

    AtomJavaTypeSystem ts = (AtomJavaTypeSystem)c.typeSystem();

    String failure = "FAILURE_";
    if(ts.isSubtype(type, ts.Object()))
      failure += "OBJECT";
    else if(ts.equals(type, ts.Boolean()))
      failure += "BOOLEAN";
    else if(ts.equals(type, ts.Double()) ||
            ts.equals(type, ts.Float()))
      failure += "REAL";
    else if(ts.equals(type, ts.Char()))
      failure += "CHAR";
    else if(ts.equals(type, ts.Void()))
      failure += "VOID";
    else
      failure += "INTEGER";

    if(exprs.size() == 1 && exprs.get(0) instanceof NullLit)
      exprs = AJS.List
        (Cast(c, CanonicalTypeNode(c, ts.Object()), (Expr)exprs.get(0)));
    
    Call fail = super.Call(cgp, CanonicalTypeNode(c, ts.Failure()), failure, 
                           AJS.cons(StringLit(c, s), exprs));
    MethodInstance mi = /* java 1.5 varargs hack */
      (MethodInstance) ts.Failure().methodsNamed(failure).get(0);
    fail = (Call) fail.type(mi.returnType());
    fail = fail.methodInstance(mi);
    return fail;
  }
  
  public Call currentThread(Context c) {
    ClassType athread = ((AtomJavaTypeSystem)c.typeSystem()).AThread();
    return Call(c, CanonicalTypeNode(c, athread), "currentThread",
                Collections.EMPTY_LIST);
  }
  public Field threadLocalField(Context c, String fieldName) {
    ClassType athread = ((AtomJavaTypeSystem)c.typeSystem()).AThread();
    return Field(c, athread, currentThread(c), fieldName);
  }
  public Throw doRollBack(Context c) {
    ClassType rollbackT = ((AtomJavaTypeSystem)c.typeSystem()).AtomJavaRollBack();
    return Throw(cgp, Field(c, rollbackT, CanonicalTypeNode(c,rollbackT), 
                            "ROLLBACK"));
  }


  public Assign Assign(Context c, Expr left, Assign.Operator op, Expr right) {
    Assign a;
    if (left instanceof Local)
      a = super.LocalAssign(cgp, (Local)left, op, right);
    else if (left instanceof Field)
      a = super.FieldAssign(cgp, (Field)left, op, right);
    else if (left instanceof ArrayAccess)
      a = super.ArrayAccessAssign(cgp, (ArrayAccess)left, op, right);
    else 
      a = super.AmbAssign(cgp, left, op, right);
    return (Assign) a.type(left.type());
  }

  public Assign Assign(Context c, Expr left, Expr right) {
    return Assign(c,  left, Assign.ASSIGN, right);
  }
    
  public Binary Binary(Context c, Expr left, Binary.Operator op, 
                       Expr right, Type t) {
    return (Binary) super.Binary(cgp, left, op, right).type(t);
  }

  public Unary Unary(Context c, Unary.Operator op, Expr expr) {
    Unary n = super.Unary(cgp, op, expr);
    return (Unary)n.type(expr.type());
  }

  public CanonicalTypeNode CanonicalTypeNode(Context c, Type t) {
    return (CanonicalTypeNode) super.CanonicalTypeNode(cgp, t).type(t);
  }
    
  public New New(Context c, Expr outer, TypeNode objectType, 
                 List args, ClassBody body) {
    New n = super.New(cgp, outer, objectType, args, body);
    n = (New)n.type(objectType.type());
    try {
      ConstructorInstance ci = 
        c.typeSystem().findConstructor(objectType.type().toClass(),
                                       AJS.Expr_to_Type(args),
                                       c.currentClass());
      return n.constructorInstance(ci);
    } catch(SemanticException e) { }
    return n;
  }

  public LocalDecl LocalDecl(Context c, Flags flags, TypeNode type, String name, Expr init) {
    LocalDecl ld = super.LocalDecl(cgp, flags, type, name, init);
    return ld.type(type).localInstance
      (new LocalInstance_c(c.typeSystem(), ld.position(), 
                           flags, ld.declType(), name));
  }

  public Initializer Initializer(Context c, Flags f, Block b, 
                                 ClassType container) {
    Initializer n = super.Initializer(cgp, f, b);
    return n.initializerInstance
      (new InitializerInstance_c(c.typeSystem(), cgp, container, f));
  }
  
  public ClassBody ClassBody(Context c, List members) {
    return super.ClassBody(cgp, members);
  }

  private static int var_counter = 0;

  @Override
public LocalDecl LocalDecl(Position pos, Flags flags, TypeNode type, String name, Expr init) {
    LocalDecl n = new AtomJavaLocalDecl_c(pos, flags, type, name, init);
    n = (LocalDecl)n.ext(extFactory().extLocalDecl());
    n = (LocalDecl)n.del(delFactory().delLocalDecl());
    return n;
}  

  public LocalDecl freshLocalDecl(Context c, Flags flags, String startName, 
                                  Type type, Expr init) {
    String var = 
      "__aj_" + (startName != null ? startName : "") + "_";

    while(c.findVariableSilent(var) != null)
      var = "__aj_" + startName + "_" + var_counter++ + "_";

    LocalDecl ld = 
      LocalDecl(c, flags, CanonicalTypeNode(c, type), var, init);
    c.addVariable(ld.localInstance());
    return ld;
  }

  public Lit Zero(Context c, Type type) {
      if (type.isIntOrLess()) return super.IntLit(cgp, IntLit.INT, 0L);
      if (type.isLong()) return super.IntLit(cgp, IntLit.LONG, 0L);
      if (type.isFloat()) return super.FloatLit(cgp, FloatLit.FLOAT, 0.0F);
      if (type.isDouble()) return super.FloatLit(cgp, FloatLit.DOUBLE, 0.0);
      if (type.isBoolean()) return super.BooleanLit(cgp, false);
      return super.NullLit(cgp);
  }

  //// factory for nodes with relationships ////

  public Call Call(Context c, Receiver target, List args, MethodDecl md) {
    MethodInstance mi = md.methodInstance();
    Call call = super.Call(cgp, target, mi.name(), args);
    call = (Call) call.type(mi.returnType());
    call = (Call) call.methodInstance(mi);
    return call;
  }

  public Local Local(Context c, LocalInstance li) {
    return (Local) super.Local(cgp, li.name())
      .localInstance(li)
      .type(li.type());
  }
  
  public Local Local(Context c, LocalDecl ld) {
    LocalInstance li = ld.localInstance();
    Local l = super.Local(cgp, li.name());
    return (Local)l.localInstance(li).type(li.type());
  }
  
  public Local Local(Context c, Formal f) {
    LocalInstance li = f.localInstance();
    Local l = super.Local(cgp, li.name());
    return ((Local)l.type(li.type())).localInstance
      (new LocalInstance_c(li.typeSystem(), l.position(),
                           li.flags(), li.type(), li.name()));
  }
  
  public Field Field(Context c, FieldDecl fd, Receiver r) {
    FieldInstance fi = fd.fieldInstance();
    if(!(r.type().equals(fi.container())))
      throw new InternalCompilerError("Receiver of Field expr is not same type as FieldDecl's container");
    
    Field n = super.Field(cgp, r, fi.name());
    return ((Field)n.type(fi.type())).fieldInstance(new FieldInstance_c(fi.typeSystem(), n.position(), fi.container(), fi.flags(), fi.type(), fi.name()));
  }
  
  public Field Field(Context c, FieldDecl fd) {
    FieldInstance fi = fd.fieldInstance();
    if(fi.flags().isStatic()) {
      Field n = super.Field(cgp, CanonicalTypeNode(c, fi.container()), fi.name());
      return ((Field)n.type(fi.type())).fieldInstance(new FieldInstance_c(fi.typeSystem(), n.position(), fi.container(), fi.flags(), fi.type(), fi.name()));
    }
    
    if(!(fi.container().equals(c.currentClass())))
      throw new InternalCompilerError("Implicit Field expression where FieldDecl container is not current class in scope");
    
    Field n = super.Field(cgp, super.Special(cgp, Special.THIS, null).type(c.currentClass()), fi.name());
    return ((Field)n.type(fi.type())).fieldInstance(new FieldInstance_c(fi.typeSystem(), n.position(), fi.container(), fi.flags(), fi.type(), fi.name()));
  }
  
  // given a ClassType and needing a field you could only ever want a static field
  public Field Field(Context c, ClassType t, String name) {
    FieldInstance fi = t.fieldNamed(name);
    if(fi == null)
      throw new InternalCompilerError("could not find field named: " + name + " in class type: " + t);
    
    if(fi.flags().isStatic()) {
      Field n = super.Field(cgp, super.CanonicalTypeNode(cgp, t), name);
      return ((Field)n.type(fi.type())).fieldInstance(fi);
    }
    
    if(!(fi.container().equals(c.currentClass())))
      throw new InternalCompilerError("Implicit Field expression where FieldDecl container is not current class in scope");

    Field n = super.Field(cgp,super.Special(cgp, Special.THIS, null).type(c.currentClass()), name);
    return ((Field)n.type(fi.type())).fieldInstance(fi);
  }
  
  public Field Field(Context c, ClassType t, Receiver r, String name) {
    FieldInstance fi = t.fieldNamed(name);
    if(fi == null)
      throw new InternalCompilerError
        ("Could not find field named: " + name + " in class type: " + t);
    
    if(!(r.type().isSubtype(fi.container())))
      throw new InternalCompilerError
        ("Receiver of Field expr not same type as FieldDecl container");
    
    Field n = super.Field(cgp, r, fi.name());
    return ((Field)n.type(fi.type())).fieldInstance(fi);
  }
}
