package polyglot.ext.atomjava.ast;

import polyglot.ext.atomjava.types.*;
import polyglot.ext.atomjava.ast.*;
import polyglot.ext.atomjava.util.AJS;
import polyglot.ext.atomjava.visit.AtomJavaBuilder;
import polyglot.ext.jl.ast.*;
import polyglot.ast.*;
import polyglot.util.*;
import polyglot.types.*;
import polyglot.visit.*;

import java.util.*;

public class AtomJavaSynchronized_c extends Synchronized_c implements AtomJavaBuild {

  public AtomJavaSynchronized_c(Position pos, Expr expr, Block body) {
    super(pos, expr, body);
  }


  public NodeVisitor buildEnter(AtomJavaBuilder ab) {
    return ab;
  }

  public Node build(Node parent, AtomJavaBuilder ab) {
    // set okay to steal, etc.
    AtomJavaContext     c  = (AtomJavaContext)    ab.context();
    AtomJavaTypeSystem  ts = (AtomJavaTypeSystem) ab.typeSystem();
    AtomJavaNodeFactory nf = (AtomJavaNodeFactory)ab.nodeFactory();

    Position p = Position.COMPILER_GENERATED;
    Field robbed    = nf.threadLocalField(c,"robbed");
    Field revocable = nf.threadLocalField(c,"revocable");
    Field lock      = nf.threadLocalField(c,"lock");

    Assign robbed_false = nf.Assign(c, robbed,    nf.BooleanLit(p, false));
    Assign steal_true   = nf.Assign(c, revocable, nf.BooleanLit(p, true));
    Synchronized synch_true = 
      nf.Synchronized(c, lock, nf.Block(p, nf.Eval(p, steal_true), 
                                           nf.Eval(p, robbed_false)));
    Assign steal_false = steal_true.right(nf.BooleanLit(p, false));
    Stmt   synch_false = synch_true.body(nf.Block(p, nf.Eval(p, steal_false)));

    List stmts = new LinkedList(body.statements());
        
    if(c.inAtomicContext())
      stmts.add(0, nf.If(p, robbed, nf.doRollBack(c)));
        
    stmts.add(0, synch_false);
    return nf.Block(c, AJS.List(synch_true, 
                                nf.Synchronized(c, expr, nf.Block(c,stmts))));
  }
}
