package polyglot.ext.atomjava.ast;

import polyglot.ext.atomjava.types.*;
import polyglot.ext.atomjava.ast.*;
import polyglot.ext.atomjava.util.AJS;
import polyglot.ext.atomjava.visit.AtomJavaBuilder;
import polyglot.ext.jl.ast.*;
import polyglot.ast.*;
import polyglot.util.*;
import polyglot.types.*;
import polyglot.visit.*;

import java.util.*;

public class AtomJavaTry_c extends Try_c implements AtomJavaBuild {
  public AtomJavaTry_c(Position pos, Block tryBlock, List catchBlocks, Block finallyBlock) {
    super(pos, tryBlock, catchBlocks, finallyBlock);
  }

  public Context enterScope(Node child, Context c) {
    if(child.equals(tryBlock))
      return ((AtomJavaContext)c).pushTryCatch();

    return c;
  }
            
  public NodeVisitor buildEnter(AtomJavaBuilder ab) {
    return ab;
  }

  public Node build(Node parent, AtomJavaBuilder ab) {
    if(ab.context().inAtomicContext())
      return buildAtomic(parent, ab);
    return this;
  }

  public Node buildAtomic(Node parent, AtomJavaBuilder ab) {
    // rethrow RollBackException if necessary
    AtomJavaContext     c  = (AtomJavaContext)    ab.context();
    AtomJavaTypeSystem  ts = (AtomJavaTypeSystem) ab.typeSystem();
    AtomJavaNodeFactory nf = (AtomJavaNodeFactory)ab.nodeFactory();

    boolean need = false;
    if(finallyBlock != null)
      need = true;
    for(Iterator i = catchBlocks.iterator(); i.hasNext() && !need;)
      if(ts.descendsFrom(ts.AtomJavaRollBack(),((Catch)i.next()).catchType()))
        need = true;
    if(!need)
      return this;
      
    Position p = Position.COMPILER_GENERATED;
    Assign rollback = nf.Assign(c, nf.threadLocalField(c,"rollback"), 
                                nf.BooleanLit(p, true));
    Block body = nf.Block(p, nf.Eval(p, rollback));
    CanonicalTypeNode rollbackT = nf.CanonicalTypeNode(c,ts.AtomJavaRollBack());
    Formal formal = nf.Formal(c, Flags.NONE, rollbackT, "rb");
    Catch rb = nf.Catch(p, formal, body);
    List  cb = new LinkedList(catchBlocks);
    cb.add(0, rb);
                
    List stmts = (finallyBlock==null) ?
      new LinkedList() : new LinkedList(finallyBlock.statements());
                
    List if_stmts = new LinkedList();
    if_stmts.add(nf.Eval(p, nf.Assign(c, nf.threadLocalField(c,"rollback"),
                                      nf.BooleanLit(p, false))));
    if_stmts.add(nf.doRollBack(c));

    stmts.add(0, nf.If(p, nf.threadLocalField(c,"rollback"),
                       nf.Block(p, if_stmts), null));
     
    return nf.Try(position, tryBlock, cb, nf.Block(p,stmts));
  }
}
