package polyglot.ext.atomjava.ast;

import polyglot.ext.jl.ast.Unary_c;
import polyglot.visit.*;
import polyglot.ast.*;
import polyglot.types.*;
import polyglot.util.*;
import polyglot.ext.atomjava.main.AJOptions;
import polyglot.ext.atomjava.types.*;
import polyglot.ext.atomjava.util.*;
import polyglot.ext.atomjava.visit.*;

public class AtomJavaUnary_c 
  extends Unary_c 
  implements AtomJavaBuild {

  public AtomJavaUnary_c(Position pos, Unary.Operator op, Expr expr) {
    super(pos, op, expr);
  }

  public NodeVisitor buildEnter(AtomJavaBuilder ab) {
    return ab;
  }

  public Node build(Node parent, AtomJavaBuilder ab) {
    AJOptions options = (AJOptions) ab.job().extensionInfo().getOptions();
    if(options.weak && !ab.context().inAtomicContext())
      return buildWeak(parent, ab);
    else
      return buildStrong(parent, ab);
  }
  
  public Node buildWeak(Node parent, AtomJavaBuilder ab) {
    return this;
  }
      
  public Node buildStrong(Node parent, AtomJavaBuilder ab) {
    boolean assign_op = (op == Unary.POST_INC || op == Unary.POST_DEC || 
                         op == Unary.PRE_INC || op == Unary.PRE_DEC);

    AtomJavaContext c = ab.context();
    AtomJavaNodeFactory nf = ab.nodeFactory();

    Call call;
    
    if(expr instanceof Field && assign_op) {

      Field field = (Field) expr;
      Receiver e1 = field.target();

      Expr target = (e1 instanceof Expr) ? 
        (Expr) e1 : nf.NullLit(c);

      Assign.Operator assop = (op == Unary.POST_INC || op == Unary.PRE_INC) ? 
        Assign.ADD_ASSIGN : Assign.SUB_ASSIGN;

      call = 
        nf.Call(c, null, AJS.List(target, nf.IntLit(c, IntLit.INT, 1)),
                c.makeFieldOperatorMethod(field, assop, ab));

    } else if(expr instanceof ArrayAccess && assign_op) {

      Expr array = ((ArrayAccess)expr).array();
      Expr index = ((ArrayAccess)expr).index();
            
      Assign.Operator assop = (op == Unary.POST_INC || op == Unary.PRE_INC) ? 
        Assign.ADD_ASSIGN : Assign.SUB_ASSIGN;

      call = 
        nf.Call(c, null, AJS.List(array,index,nf.IntLit(c,IntLit.INT,1)),
                c.makeArrayOperatorMethod(array.type(), assop, ab));

    } else return this;

    if(op == Unary.POST_INC && !(parent instanceof Eval))
      return nf.Binary(c, call, Binary.SUB, 
                       nf.IntLit(c, IntLit.INT, 1), type);
    if(op == Unary.POST_DEC && !(parent instanceof Eval))
      return nf.Binary(c, call, Binary.ADD, 
                       nf.IntLit(c, IntLit.INT, 1), type);

    return call;
  }
}
