package polyglot.ext.atomjava.frontend;

import polyglot.ast.*;
import polyglot.types.*;
import polyglot.util.*;
import polyglot.frontend.*;
import polyglot.visit.*;

import java.util.*;

public class SingleVisitorJob extends Job
{
  /**
   * The job that this SinglePassJob was spawned by.
   */
  protected Job outer;
  
  /**
   * The context in which this Job was spawned.
   */
  protected Context context;
  
  /**
   * Only this pass will be performed.
   */
  protected VisitorPass pass;

  public SingleVisitorJob(ExtensionInfo lang, JobExt ext, Node ast, Context c,
                          Job outer, NodeVisitor v, Pass.ID id) {
    super(lang, ext, ast);
    
    this.context = context;
    this.outer = outer;
    this.pass = new VisitorPass(id, this, v);
    if (ast == null) {
      throw new InternalCompilerError("Null ast");
    }
    if (outer == null) {
      throw new InternalCompilerError("Null outer job");
    }
  }
  
  /**
   * Only the single pass represents this <code>Job</code>
   */
  public List getPasses() {
    return Collections.singletonList(pass);
  }
  
  public Context context() {
    return context;
  }
  
  /**
   * The <code>SourceJob</code> associated with our outer <code>Job</code>.
   */
  public SourceJob sourceJob() {
    return outer.sourceJob();
  }
}
