package polyglot.ext.atomjava.types;

import polyglot.ast.*;
import polyglot.types.*;
import polyglot.ext.jl.types.*;

import polyglot.util.InternalCompilerError;

import polyglot.ext.atomjava.util.AJS;
import polyglot.ext.atomjava.ast.*;
import polyglot.ext.atomjava.visit.*;

import java.util.*;

public class AtomJavaContext_c extends Context_c implements AtomJavaContext {

  protected boolean atomic_context;
  protected boolean trycatch_context;
  protected Map<String, MethodDecl> operator_helpers;

  protected boolean atomic_block;

  public AtomJavaContext_c(AtomJavaTypeSystem ts) {
    super(ts);
    atomic_context = false;
    trycatch_context = false;
  }

  public List getLocalVars() {
    return getLocalVars(new LinkedList());
  }

  public List getLocalVars(List locals) {
    if (vars != null) {
      Iterator keys = vars.keySet().iterator();
      while(keys.hasNext()) {
        VarInstance vi = (VarInstance)vars.get((String)keys.next());
        if(vi instanceof LocalInstance)
          locals.add(vi);
      }
    }

    if (outer != null) {
      return ((AtomJavaContext)outer).getLocalVars(locals);
    }

    return locals;
  }    

  public Context pushTryCatch() {
    if(trycatch_context) return this;
    AtomJavaContext_c c = (AtomJavaContext_c) push();
    c.atomic_context = atomic_context;
    c.trycatch_context = true;
    return c;
  }
  public boolean inTryCatchContext() {
    return trycatch_context;
  }

  public Context pushAtomicBlock() {
    if(atomic_block) return this;
    AtomJavaContext_c c = (AtomJavaContext_c) push();
    c.atomic_block = true;
    return c;
  }
  public boolean inAtomicBlock() {
    return atomic_block;
  }
  public Context pushAtomic() {
    if(atomic_context) return this;
    AtomJavaContext_c c = (AtomJavaContext_c) push();
    c.atomic_context = true;
    return c;
  }
  public Context ignoreAtomic() {
    AtomJavaContext_c c = (AtomJavaContext_c) push();
    c.atomic_context = false;
    return c;
  }
  public boolean inAtomicContext() {
    return atomic_context;
  }

  private AtomJavaContext_c getClassContext() {
    if(this.kind == CLASS) {
      if(operator_helpers == null)
        operator_helpers = new HashMap<String, MethodDecl>();
      if(failure_methods == null)
        failure_methods = new HashMap<String, MethodDecl>();
      return this;
    }
    return ((AtomJavaContext_c)outer).getClassContext();
  }
  private boolean needOperatorHelper(String method) {
    return !getClassContext().operator_helpers.containsKey(method);
  }
  private void putOperatorHelper(String method, MethodDecl md) {
    getClassContext().operator_helpers.put(method, md);
  }
  private MethodDecl getOperatorHelper(String method) {
    return getClassContext().operator_helpers.get(method);
  }
  private Map operatorHelpers() {
    return getClassContext().operator_helpers;
  }

  public MethodDecl makeArrayOperatorMethod(Type array_type, 
                                            Assign.Operator assop,
                                            AtomJavaBuilder ab) {

    AtomJavaNodeFactory nf = ab.nodeFactory();

    Type    base_type = array_type.toArray().base();

    Binary.Operator op = AJS.ConvertOperator(assop);
    String n = AJS.OperatorString(op) + "_assignarray_" + 
      base_type.toString().replace('.','_');
    String method_name = operatorHelperName(n);

    if(needOperatorHelper(method_name)) {
      AtomJavaTypeSystem  ts  = ab.typeSystem();

      Formal a = nf.Formal(this, Flags.NONE, 
                           nf.CanonicalTypeNode(this, array_type), "a");
      Formal i = nf.Formal(this, Flags.NONE, 
                           nf.CanonicalTypeNode(this, ts.Int()), "i");
      Formal v = nf.Formal(this, Flags.NONE, 
                           nf.CanonicalTypeNode(this, base_type), "v");

      Call get = nf.Call(this, nf.CanonicalTypeNode(this, ts.AArray()),
                         arrayElementGetterName(base_type),
                         AJS.List(nf.Local(this, a), nf.Local(this, i)));

      Binary b = nf.Binary(this, get, op, nf.Local(this, v), base_type);

      Call set = nf.Call(this, nf.CanonicalTypeNode(this, ts.AArray()),
                         arrayElementSetterName(base_type),
                         AJS.List(nf.Local(this, a), nf.Local(this, i), b));

      Return ret = nf.Return(ab.context(), set);

      MethodDecl md = nf.MethodDecl(this, Flags.PRIVATE.set(Flags.STATIC),
                                    base_type, method_name,
                                    AJS.List(a, i, v), AJS.List(ret));
      putOperatorHelper(method_name, md);
      ab.addClassChange(md);
      return md;
    }
    return getOperatorHelper(method_name);
  }

  public MethodDecl makeFieldOperatorMethod(Field field, 
                                            Assign.Operator assop,
                                            AtomJavaBuilder ab) {

    AtomJavaNodeFactory nf = ab.nodeFactory();

    Type target_type = field.target().type();
    Type field_type = field.type();

    Binary.Operator op = AJS.ConvertOperator(assop);
    // TO DO: this mangling could lead to conflicts
    String n = AJS.OperatorString(op) + "_assign_" + 
      target_type.toString().replace('.','_') + "_" + field.name();
    String method_name = operatorHelperName(n);

    if(needOperatorHelper(method_name)) {
      Formal x = 
        nf.Formal(this, Flags.NONE, 
                  nf.CanonicalTypeNode(this, target_type), "x");
      Formal y = 
        nf.Formal(this, Flags.NONE, 
                  nf.CanonicalTypeNode(this, field_type), "y");

      Call get = nf.Call(this, nf.CanonicalTypeNode(this, target_type),
                         fieldGetterName(field.name()), 
                         AJS.List(nf.Local(this, x)));

      Binary b = nf.Binary(this, get, op, nf.Local(this, y), field_type);

      // if the type of the class containing the field is an innerClass
      // then look for the first outer class that can have static members
      // and make a call to that class
      ClassType ct = (ClassType) target_type;
      String mangled_name = field.name();
      for(; ct.isInnerClass(); ct = ct.outer())
	  mangled_name = ct.name() + "_" + mangled_name;


      Call set = nf.Call(this, nf.CanonicalTypeNode(this, ct),
                         fieldSetterName(mangled_name), 
                         AJS.List(nf.Local(this, x), b));

      Return ret = nf.Return(ab.context(), set);

      MethodDecl md = nf.MethodDecl(this, Flags.PRIVATE.set(Flags.STATIC),
                                    field_type, method_name,
                                    AJS.List(x, y), AJS.List(ret));
      putOperatorHelper(method_name, md);
      ab.addClassChange(md);
      return md;
    }
    return getOperatorHelper(method_name);
  }

  /* TODO: FAILURE METHODS WILL MUCH MORE EASILY LEAD TO NAME CLASHES!! */

  protected Map<String, MethodDecl> failure_methods;

  private boolean needFailureMethod(String method) {
    return !getClassContext().failure_methods.containsKey(method);
  }
  private void putFailureMethod(String method, MethodDecl md) {
    getClassContext().failure_methods.put(method, md);
  }
  private MethodDecl getFailureMethod(String method) {
    return getClassContext().failure_methods.get(method);
  }

  public MethodDecl makeThrowingFailureMethod(Call call, 
                                              List throwTypes,
                                              AtomJavaBuilder ab) {
    String method_name = "__aj_Failure_";
    if(call.target().type().isArray())
      method_name += call.target().type().toArray().ultimateBase().toString() 
        + "_" + call.target().type().toArray().dims();
    else
      method_name += call.target().type().toClass().fullName().replace('.','_');

    method_name += "_" + call.name();

    for(Iterator i = throwTypes.iterator(); i.hasNext();)
      method_name += "_" + ((TypeNode)i.next()).type().toClass().fullName().replace('.','_');

    if(needFailureMethod(method_name)) {
      AtomJavaContext     c  = ab.context();
      AtomJavaTypeSystem  ts = ab.typeSystem();
      AtomJavaNodeFactory nf = ab.nodeFactory();
      
      List formals = new LinkedList();
      /* add the receiver */
      formals.add(0, nf.Formal(this, Flags.NONE,
                               nf.CanonicalTypeNode(this, call.target().type()),
                               "e0"));

      for(Iterator i = AJS.Expr_to_Type(call.arguments()).iterator(); 
          i.hasNext();) {
        Formal f = nf.Formal(this, Flags.NONE, 
                             nf.CanonicalTypeNode(this, (Type)i.next()), 
                             "e" + formals.size());
        formals.add(f);
      }
      
      New n = 
        nf.New(this, null, 
               nf.CanonicalTypeNode(this, ts.NoSuchMethodError()),
               Collections.EMPTY_LIST, null);
      
      Throw t = nf.Throw(this, n);
      
      // name the method based on the original call
      MethodDecl md = 
        nf.MethodDecl(this, Flags.PRIVATE.set(Flags.STATIC),
                      nf.CanonicalTypeNode(this, call.methodInstance().returnType()),
                      method_name, formals, throwTypes, 
                      nf.Block(this, AJS.List(t)));
      putFailureMethod(method_name, md);
      ab.addClassChange(md);
    }
    return getFailureMethod(method_name);
  }
  
  public String operatorHelperName(String op) {
    if(!inAtomicContext())
      return "__aj_operator_" + op;
    return "__aj_operator_atomic_" + op;
  }

  public String atomicMethodName(String orig) {
    return "__aj_atomic_" + orig;
  }

  public String fieldSetterName(String field) {
    if(!inAtomicContext())
      return "__aj_set_" + field;
    return "__aj_set_atomic_" + field;
  }
    
  public String fieldGetterName(String field) {
    return "__aj_get_" + field;
  }
    
  public String fieldUndoerName(String field) {
    return "__aj_undo_" + field;
  }
    
  public String atomicFormalName() {
    return "__aj_0";
  }

  public String atomicConstructorFieldName() {
    return "acons";
  }
    
  public String arrayElementGetterName(Type type) {
    return "get_" + (type.isPrimitive() ? type.toString() : "T");
  }

  public String arrayElementSetterName(Type type) {
    if(!inAtomicContext())
      return "set_" + (type.isPrimitive() ? type.toString() : "T");
    return "set_atomic_" + (type.isPrimitive() ? type.toString() : "T");
  }
}
