package polyglot.ext.atomjava.visit;

import polyglot.ast.*;
import polyglot.ext.atomjava.types.*;
import polyglot.ext.atomjava.util.*;
import polyglot.ext.atomjava.ast.*;
import polyglot.frontend.*;
import polyglot.types.*;
import polyglot.visit.*;
import polyglot.util.*;

import polyglot.ext.atomjava.ExtensionInfo;

// rewrites ClassDecls to extend the correct class, AObject, AThread
public class ExtendsRewriter extends AJContextVisitor {

  public ExtendsRewriter(Job j, TypeSystem ts, NodeFactory nf) {
    super(j, ts, nf);
  }

  public Node leaveCall(Node old, Node n, NodeVisitor v) 
    throws SemanticException {
    n = super.leaveCall(old, n, v);

    if(!(n instanceof ClassDecl))
      return n;

    ClassDecl cd = (ClassDecl) n;
    
    AtomJavaContext     c  = context();
    AtomJavaNodeFactory nf = nodeFactory();
    AtomJavaTypeSystem  ts = typeSystem();

     if(cd.flags().isInterface())
       return n;

    if(cd.superClass() == null || 
       ts.equals(cd.superClass().type(), ts.Object()))
      return update(cd, ts.AObject(), c, nf, ts);

    if(ts.equals(cd.superClass().type(), ts.Thread())) 
      return update(cd, ts.AThread(), c, nf, ts);

    return n;
  }

  private Node update(ClassDecl cd, final ClassType ct, 
                      AtomJavaContext c, 
                      AtomJavaNodeFactory nf,
                      AtomJavaTypeSystem ts) {
    cd = cd.superClass(nf.CanonicalTypeNode(c, ct));
    cd.type().superType(ct);
    // declare the type update
    ts.addTypeUpdate(cd.type());
    
    // convert super calls (constructor calls and method calls)
    ExtensionInfo ext = (ExtensionInfo) job().extensionInfo();
    Job job = ext.spawnJob(c, cd.body(), job(), new HaltingVisitor() {
        public NodeVisitor enter(Node parent, Node n) {
          if(n instanceof ClassDecl)
            return ((HaltingVisitor)super.enter(parent, n)).bypassChildren(n);
          return super.enter(parent, n);
        }
        public Node leave(Node parent, Node old, Node n, NodeVisitor v) {
          if(n instanceof ConstructorCall) {
            ConstructorCall cc = (ConstructorCall)n;
            if(cc.kind() == ConstructorCall.THIS)
              return cc;
            ConstructorInstance ci = cc.constructorInstance();
            ci = ci.container(ct);
            cc = cc.constructorInstance(ci);
            return cc;
          } else if (n instanceof Call) {
            Call call = (Call)n;
            Receiver r = call.target();
            if(!(r instanceof Special) || ((Special)r).kind() == Special.THIS)
              return call;
            MethodInstance mi = call.methodInstance();
            mi = mi.container(ct);
            call = call.methodInstance(mi);
            call = call.target(((Special)r).type(ct));
            return call;
          } else return super.leave(parent, old, n, v);
        }
      }, new Pass.ID("AJ-spawn-convert-ConstructorCall-" + 
                     cd.type().fullName() + "-extends-" + ct.fullName()));
    cd = cd.body((ClassBody)job.ast());
    return cd;    
  }

}
