package polyglot.ext.atomjava.visit;

import polyglot.ast.*;
import polyglot.ext.atomjava.types.*;
import polyglot.ext.atomjava.ast.*;
import polyglot.frontend.Job;
import polyglot.types.*;
import polyglot.visit.*;
import polyglot.util.*;


// changes all typed expressions to have types that reflect 
// the current types being used by the typesystem
public class TypedUpdate extends AtomJavaContextVisitor {

  public TypedUpdate(Job j, TypeSystem ts, NodeFactory nf) {
    super(j, ts, nf);
  }

  public Node leaveCall(Node old, Node n, NodeVisitor v) 
    throws SemanticException {

    n = super.leaveCall(old, n, v);

    AtomJavaTypeSystem ts = typeSystem();

    /* declaration's with instances */

    if(n instanceof Initializer)
      n = ((Initializer)n).initializerInstance
        (ts.updateInitializerInstance(((Initializer)n).initializerInstance()));
    
    if(n instanceof MethodDecl)
      n = ((MethodDecl)n).methodInstance
        (ts.updateMethodInstance(((MethodDecl)n).methodInstance()));
    
    if(n instanceof ConstructorDecl)
      n = ((ConstructorDecl)n).constructorInstance
        (ts.updateConstructorInstance(((ConstructorDecl)n).constructorInstance()));

    if(n instanceof FieldDecl)
      n = ((FieldDecl)n).fieldInstance
        (ts.updateFieldInstance(((FieldDecl)n).fieldInstance()));

    if(n instanceof LocalDecl)
      n = ((LocalDecl)n).localInstance
        (ts.updateLocalInstance(((LocalDecl)n).localInstance()));

    if(n instanceof Formal)
      n = ((Formal)n).localInstance
        (ts.updateLocalInstance(((Formal)n).localInstance()));

    /* expression's with instances */

    if(n instanceof Field)
      n = ((Field)n).fieldInstance
        (ts.updateFieldInstance(((Field)n).fieldInstance()));
    
    if(n instanceof Local)
      n = ((Local)n).localInstance
        (ts.updateLocalInstance(((Local)n).localInstance()));

    if(n instanceof New)
      n = ((New)n).constructorInstance
        (ts.updateConstructorInstance(((New)n).constructorInstance()));
    
    if(n instanceof Call)
      n = ((Call)n).methodInstance
        (ts.updateMethodInstance(((Call)n).methodInstance()));

    if(n instanceof ConstructorCall)
      n = ((ConstructorCall)n).constructorInstance
        (ts.updateConstructorInstance(((ConstructorCall)n).constructorInstance()));
    
    if(!(n instanceof Typed))
      return n;

    if(n instanceof Expr)
      return ((Expr)n).type(ts.updateType(((Typed)n).type()));

    if(n instanceof TypeNode)
      return ((TypeNode)n).type(ts.updateType(((Typed)n).type()));    

    throw new InternalCompilerError
      ("Typed AST node is not Expr or TypeNode!", n.position());
  }
}
